@@ -2,13 +2,104 @@
 # vim: set fileencoding=utf-8 :
 
 import re
+import shlex
 import time
 import xmpp
+import xmpp.simplexml
+import sqlalchemy.ext.declarative
+import sqlalchemy.orm.exc
+import sqlalchemy.orm
+import sqlalchemy
+from optparse import OptionParser
 from collections import defaultdict
 from random import randint as rand
 from math import ceil as ceiling
 from quinoa import Bot
 
+# ~~~~~~~ Special Option Parsing
+
+class JabberOptionParser(OptionParser):
+    def exit(self, *args, **kwargs):
+        raise ValueError, "something made optparse exit."
+    def error(self, *args, **kwargs):
+        raise ValueError, "incorrect option or argument"
+
+# ~~~~~~~ DATABASE definition
+path = '/home/kit/Desktop/hg-repos/quinoa/'
+#path = ''
+engine = sqlalchemy.create_engine('sqlite:///%styche.db' % path)
+Session = sqlalchemy.orm.sessionmaker(bind=engine)
+
+Base = sqlalchemy.ext.declarative.declarative_base()
+class User(Base):
+    __tablename__ = 'users'
+    jid = sqlalchemy.Column(sqlalchemy.Unicode, primary_key=True)
+    batsignal = sqlalchemy.Column(sqlalchemy.Boolean)
+    points = sqlalchemy.Column(sqlalchemy.Integer, default=0)
+    def __init__(self, jid):
+        self.jid = jid
+        self.batsignal = True
+        self.points = 0
+    def __unicode__(self):
+        aliases = ' a.k.a. '.join(unicode(a) for a in self.aliases)
+        if not aliases:
+            aliases = 'No-name'
+        s = 's'
+        if self.points == 1 or self.points == -1:
+            s = ''
+        return "%s (%s), %d point%s" % (aliases, self.jid, self.points, s)
+    __str__ = __unicode__
+
+class JidAlias(Base):
+    __tablename__ = 'jid_aliases'
+    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
+    additional_jid = sqlalchemy.Column(sqlalchemy.Unicode)
+    primary_jid = sqlalchemy.Column(sqlalchemy.Unicode, \
+                             sqlalchemy.ForeignKey('users.jid'))
+    user = sqlalchemy.orm.relation(User, \
+            backref=sqlalchemy.orm.backref('jid_aliases', order_by=id))
+    def __init__(self, additional_jid):
+        self.additional_jid = additional_jid
+
+class Alias(Base):
+    __tablename__ = 'aliases'
+    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True)
+    user_jid = sqlalchemy.Column(sqlalchemy.Unicode, \
+                             sqlalchemy.ForeignKey('users.jid'))
+    name = sqlalchemy.Column(sqlalchemy.Unicode)
+    user = sqlalchemy.orm.relation(User, \
+            backref=sqlalchemy.orm.backref('aliases', order_by=id))
+    def __init__(self, name):
+        self.name = name
+    def __unicode__(self):
+        return self.name
+    __str__ = __unicode__
+
+class PendingConnection(Base):
+    __tablename__ = 'pending_connections'
+    primary_jid = sqlalchemy.Column(sqlalchemy.Unicode)
+    additional_jid = sqlalchemy.Column(sqlalchemy.Unicode, primary_key=True)
+    def __init__(self, primary_jid, additional_jid):
+        self.primary_jid = primary_jid
+        self.additional_jid = additional_jid
+    def process(self, msg):
+        if msg.getBody().strip('!').lower() == 'y':
+            ret = True
+            session = Session()
+            try:
+                user = session.query(User).filter_by(jid=self.primary_jid).one()
+            except sqlalchemy.orm.exc.NoResultFound:
+                return False
+            session.add(user)
+            user.jid_aliases.append(JidAlias(self.additional_jid))
+            session.commit()
+        else:
+            ret = False
+        return ret
+
+Base.metadata.create_all(engine)
+# ~~~~~~~ END DATABASE
+
 def owod(dice, diff, spec=False, will=False):
     def norm_roll(dice):
         return sorted(rand(1, 10) for x in xrange(dice))
@@ -38,9 +129,35 @@ def owod(dice, diff, spec=False, will=False):
         succs = max(0, succs - ones)
     return "%s (%s)" % (succs, ', '.join(map(str, ret)))
 
+def nwod(dice, again=10, rote=False):
+    threshold = 8
+    chance = False
+    if dice < 1:
+        dice = 1
+        threshold = 10
+        again = 10
+        chance = True
+    pool = sorted(rand(1, 10) for x in xrange(dice))
+    if chance and pool[0] == 1:
+        return "Critical failure! (%s)" % ', '.join(str(x) for x in pool)
+    if rote:
+        for i, val in enumerate(pool):
+            if val < threshold:
+                pool[i] = rand(1, 10)
+    addendum = pool
+    while sum(i >= again for i in addendum):
+        addendum = sorted(rand(1, 10) for x in
+                   range(sum(i >= again for i in addendum)))
+        pool.extend(addendum)
+    successes = sum(i >= threshold for i in pool)
+    if not successes:
+        return "Failure (%s)" % ', '.join(str(x) for x in pool)
+    return "Success %s (%s)" % (successes, ', '.join(str(x) for x in pool))
+
 def exalted(dice):
     pool = sorted(rand(1, 10) for x in xrange(dice))
     succs = sum(x >= 7 for x in pool)
+    succs += pool.count(10)
     ones = pool.count(1)
     if not succs and ones:
         return "Botch (x%s)" % ones
@@ -199,29 +316,23 @@ def hande(heaven, earth, passing_grade=1):
     pool = [rand(1, 10) for i in range(heaven)]
     num_rerolls_used = 0
     if earth < 0:
-        while num_rerolls_used < abs(earth):
-            for i, v in enumerate(pool):
-                if v >= 7:
-                    pool[i] = rand(1, 10)
-                    num_rerolls_used += 1
-                    break
-            if not sum(v >= 7 for v in pool):
-                break
-    else:
-        while num_rerolls_used < earth:
-            for i, v in enumerate(pool):
-                if v < 7:
-                    pool[i] = rand(1, 10)
-                    num_rerolls_used += 1
-                    break
-            if not sum(v < 7 for v in pool):
+        earth = -earth
+    while num_rerolls_used < earth:
+        for i, v in enumerate(pool):
+            if v < 7:
+                pool[i] = rand(1, 10)
+                num_rerolls_used += 1
                 break
+        if not sum(v < 7 for v in pool):
+            break
     hits = sum(x >= 7 for x in pool) + pool.count(10)
-    if hits >= passing_grade:
-        ret = "Success %d (%s)" % \
-                (hits + 1 - passing_grade, ', '.join(str(x) for x in pool))
-    else:
-        ret = "Failure %d (%s)" % \
+    if hits == passing_grade:
+        ret = "Pass (%s)" % (', '.join(str(x) for x in pool))
+    if hits > passing_grade:
+        ret = "Pass +%d (%s)" % \
+                (hits - passing_grade, ', '.join(str(x) for x in pool))
+    if hits < passing_grade:
+        ret = "Fail -%d (%s)" % \
                 (passing_grade - hits, ', '.join(str(x) for x in pool))
     if num_rerolls_used < abs(earth):
         val = abs(earth) - num_rerolls_used
@@ -235,15 +346,175 @@ def generic(num, size):
     return ', '.join(map(str, (rand(1, size) for x in xrange(num))))
 
 class DiceBot(Bot):
+    def __init__(self, *args, **kwargs):
+        Bot.__init__(self, *args, **kwargs)
+        self.mode = 'nwod'
     def on_connect(self):
         join = xmpp.protocol.Message()
         join.setBody('join ooc@rooms.transneptune.net')
         self.join(join)
     def register_commands(self):
-        self.commands[r'mode\b'] = self.mode
-        self.commands[r'roll\b'] = self.roll
-        self.commands[r'init\b'] = self.initiative
+        self.commands[r'%s' % '|'.join(SOUND_EFFECTS)] = self.sound_effects
+        self.commands[r'[Mm]ode\b'] = self.mode
+        self.commands[r'[Rr]oll\b'] = self.roll
+        self.commands[r'[Ii]nit\b'] = self.initiative
         self.commands[r'(?i)%s' % '|'.join(MEMES)] = self.meme
+        self.commands[r'[Aa]ccount\b'] = self.remember_me
+        self.commands[r'[Ww]ho is\b'] = self.who_is
+        self.commands[r'(?i)batsignal\??$'] = self.batsignal
+        self.commands[r'!\b'] = self.confirm_user
+        self.commands[r'[Gg]ive\b'] = self.points
+    def points(self, msg):
+        """Give points to someone on the batsignal.  Usage: give USER X points"""
+        args = msg.getBody()
+        try:
+            cmd, un, number, points = args.split()
+            number = int(number)
+        except:
+            return
+        session = Session()
+        try:
+            user = session.query(Alias).filter(Alias.name.like(un)).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            return "Who's that?"
+        user = user.user
+        session.add(user)
+        user.points += number
+        session.commit()
+        if user.points > 9000:
+            return "WHAT?  OVER 9000!?"
+        return "OK."
+    def remember_me(self, msg):
+        """Usage: account [-q] space separated aliases "with quotes for multiword aliases"
+
+        Options:
+            -q: do not subscribe to the BATSIGNAL
+            -j: interpret aliases as other JIDs to associate with this user"""
+        if msg.getType() == 'groupchat':
+            return "Please use this in a private chat."
+        args = msg.getBody()
+        try:
+            cmd, args = args.split(None, 1)
+        except:
+            return
+        account = msg.getFrom().getNode() + '@' + msg.getFrom().getDomain()
+        parser = JabberOptionParser()
+        parser.add_option("-q", action="store_true", dest="quiet",
+                help="Ignore the BATSIGNAL.")
+        parser.add_option("-j", action="store_true", dest="jid",
+                help="Interpret args as JIDs.")
+        opts, args = parser.parse_args([s.decode('utf-8') for s in \
+                shlex.split(args.encode('utf-8'))])
+        # if another JID is specified, create a State, add it to self.convos,
+        # and do a special call to self.send, to alert the other JID to reply
+        # with a confirmation or denial.  Then, return with a note about the
+        # behavior to expect.
+        session = Session()
+        try:
+            user = session.query(User).filter_by(jid=account).one()
+        except sqlalchemy.orm.exc.NoResultFound:
+            user = User(account)
+        session.add(user)
+        if opts.quiet:
+            user.batsignal = False
+        else:
+            user.batsignal = True
+        if args and not opts.jid:
+            user.aliases = [Alias(x) for x in args]
+            ret = unicode(user)
+        if args and opts.jid:
+            confirm_msg = "%s claims you are another Jabber identity of " \
+                 "theirs; if this is false, ignore this message or reply " \
+                 "with !n.  If it is true, reply with !y." \
+                 % (msg.getFrom().getNode() + "@" + msg.getFrom().getDomain())
+            from_ = msg.getFrom()
+            from_.setResource('')
+            for jid in args:
+                try:
+                    pend = session.query(PendingConnection) \
+                        .filter_by(additional_jid=jid).one()
+                except:
+                    pend = PendingConnection(unicode(from_), jid)
+                session.add(pend)
+                self._send(jid, confirm_msg, 'chat')
+            ret = "Message sent to other identity."
+        session.commit()
+        return ret
+    def who_is(self, msg):
+        """Ask who someone is, by alias or email."""
+        args = msg.getBody().strip().rstrip('?').strip()
+        try:
+            who, is_, args = args.split(None, 2)
+        except:
+            return
+        session = Session()
+        ret = []
+        if args.lower() == "john galt":
+            return "Fuck Ayn Rand."
+        for v in session.query(User).all():
+            if args.lower() == v.jid.lower():
+                ret.append(v)
+            if args.lower() in (x.name.lower() for x in v.aliases):
+                ret.append(v)
+            if args.lower() in (x.additional_jid.lower() \
+                    for x in v.jid_aliases):
+                ret.append(v)
+        return '\n'.join(unicode(x) for x in ret)
+    def batsignal(self, msg):
+        """Call for the troops!"""
+        if msg.getType() != 'groupchat':
+            return "Only use this from a Jabber room."
+        session = Session()
+        if msg.getBody().endswith("?"):
+            ret = []
+            for u in session.query(User).all():
+                if u.batsignal:
+                    ret.append(u)
+            return '\n'.join(unicode(x) for x in ret)
+        room = msg.getFrom()
+        room.setResource('')
+        for u in session.query(User).all():
+            if u.batsignal:
+                self.invite(room, xmpp.protocol.JID(u.jid))
+    def user_in_room(self, room, jid):
+        """Checks a room for a JID's presence.  If the room is anonymous,
+        returns False"""
+        # TODO make this fo' real.
+        return False
+    def invite(self, room, jid):
+        """Invites a JID to a room, if that JID is not already there."""
+        if self.user_in_room(room, jid):
+            return
+        msg = xmpp.protocol.Message(to=jid, frm=self.jid)
+        body = xmpp.simplexml.Node(tag='x',
+                                   attrs={
+                                    'xmlns': 'jabber:x:conference',
+                                    'jid': room,
+                                   })
+        msg.addChild(node=body)
+        self.conn.send(msg)
+    def confirm_user(self, msg):
+        """Creates a new user connection, or destroys a pending one, based on
+        input."""
+        additional = msg.getFrom()
+        additional.setResource('')
+        additional = unicode(additional)
+        session = Session()
+        try:
+            pending = session.query(PendingConnection) \
+                    .filter_by(additional_jid=additional).first()
+        except KeyError:
+            return "Sorry, something went wrong.  " \
+                   "Please contact kit@transneptune.net"
+        try:
+            approved = pending.process(msg)
+        except:
+            return "Are you IMing me from the right account?"
+        session.delete(pending)
+        session.commit()
+        if approved:
+            return "Thank you!"
+        return "Sorry to trouble you."
     def mode(self, msg):
         """Set or view the bot's current game mode.
             * mode
@@ -254,6 +525,7 @@ class DiceBot(Bot):
                 if <value> in possible modes, sets mode to that value."""
         modes = set([
                 "owod",
+                "nwod",
                 "exalted",
                 "btvs",
                 "allflesh",
@@ -281,6 +553,8 @@ class DiceBot(Bot):
         """There are a number of ways to roll dice.  In all cases, replace # with one or more numerals, and all elements in parentheses are optional:
             * oWoD: roll # at # (s) (w)
                 pool size, difficulty, (specialized?) (willpower spent?)
+            * nWoD: roll # (#) (r)
+                pool size, (roll again threshold?), (rote?)
             * Exalted: roll #
                 pool size
             * Buffy the Vampire Slayer: roll #
@@ -321,6 +595,21 @@ class DiceBot(Bot):
                 except ValueError, e:
                     return "Bad value: %s" % e
                 return owod(dice, diff, spec, will)
+        if self.mode == "nwod":
+            _nwod = re.compile(r'^(\d+)( \d+)?( r)?$')
+            if _nwod.search(args):
+                dice, again, rote = _nwod.search(args).groups()
+                try:
+                    dice = int(dice)
+                except ValueError, e:
+                    return "Bad value: %s" % e
+                try:
+                    again = int(again.strip())
+                except ValueError, e:
+                    again = 10
+                except AttributeError, e:
+                    again = 10
+                return nwod(dice, again, rote)
         if self.mode == "exalted":
             _exalted = re.compile(r'^(\d+)$')
             if _exalted.search(args):
@@ -439,7 +728,7 @@ class DiceBot(Bot):
             return generic(num, size)
         if _rick.search(args):
             time.sleep(rand(1, 3))
-            return "http://tinyurl.com/2msmwl"
+            return "http://is.gd/czLKl"
     def initiative(self, msg):
         """Roll initiative.
         * oWoD: init (name:value)*
@@ -529,6 +818,20 @@ class DiceBot(Bot):
             return "Your mother's lipstick."
         if re.search(MEMES[12], args, re.I):
             return "YA RLY."
+        if re.search(MEMES[13], args, re.I):
+            return "Just like that?"
+        if re.search(MEMES[14], args, re.I):
+            return "The other half is bullets."
+        if re.search(MEMES[15], args, re.I):
+            return "-ang Clan ain't nuthin' ta fuck wit'!"
+    def sound_effects(self, msg):
+        args = msg.getBody()
+        if re.search(SOUND_EFFECTS[0], args, re.I):
+            return "http://instantrimshot.com"
+        if re.search(SOUND_EFFECTS[1], args, re.I):
+            return "http://sadtrombone.com"
+        if re.search(SOUND_EFFECTS[2], args, re.I):
+            return "http://instantcrickets.com"
 
 MEMES = [r'(spartans(!|,)\s+what is your profession\?)',
         r'(tyche!\s+what is your profession\?)',
@@ -543,12 +846,19 @@ MEMES = [r'(spartans(!|,)\s+what is your profession\?)',
         r"(i'm not left( |-)handed(!|\.))",
         r'(what do you have on under (that( kilt)?|there)\?)',
         r'(o rly\??)',
+        r'(when i move[,:]? you move)',
+        r'((and )?knowing is half the battle\.?)',
+        r'(w00(t|7))',
         ]
 
+SOUND_EFFECTS = [r'\*rimshot\*',
+                 r'\*(sad )?trombone\*',
+                 r'\*crickets\*'
+                ]
+
 #~~~~~~~~~~~~~~~~~~~~~~ Run the bot.
 
 if __name__ == "__main__":
     b = DiceBot('test@transneptune.net', 'Tyche', '^^password^^')
-    b.mode = "owod"
     b.serve()
 
